import argparse
import os
from dataclasses import dataclass
from typing import Optional

from _pytest.config import UsageError

from pytest_sosu.logging import get_struct_logger
from pytest_sosu.webdriver import WebDriverUrlData

logger = get_struct_logger(__name__)


@dataclass
class SosuConfig:
    username: str
    access_key: str
    region: Optional[str]
    webdriver_url_data: WebDriverUrlData

    @property
    def webdriver_url_data_with_credentials(self) -> WebDriverUrlData:
        if not self.webdriver_url_data.has_credentials:
            return self.webdriver_url_data.with_credentials(
                self.username, self.access_key
            )
        return self.webdriver_url_data


def build_sosu_config(args: argparse.Namespace, env: os._Environ) -> SosuConfig:
    logger.debug("build_sosu_config", args=args, env=env)
    username = args.sosu_username or env.get("SAUCE_USERNAME")
    access_key = args.sosu_access_key or env.get("SAUCE_ACCESS_KEY")

    if not username:
        raise UsageError("--sosu-username or SAUCE_USERNAME are not provided")

    if not access_key:
        raise UsageError("--sosu-access-key or SAUCE_ACCESS_KEY are not provided")

    region: Optional[str] = args.sosu_region or env.get("SAUCE_REGION")
    webdriver_url = args.sosu_webdriver_url or env.get("SAUCE_WEBDRIVER_URL")

    if not webdriver_url:
        host = get_host_by_region(region)
        webdriver_url_data = WebDriverUrlData(host=host)
    else:
        try:
            webdriver_url_data = WebDriverUrlData.from_url(webdriver_url)
        except ValueError:
            raise UsageError("Invalid WebDriver URL") from None

    return SosuConfig(
        username=username,
        access_key=access_key,
        region=region,
        webdriver_url_data=webdriver_url_data,
    )


def get_host_by_region(region: Optional[str]) -> str:
    if not region or region == "global":
        return "ondemand.saucelabs.com"
    host_region = REGION_MAP.get(region, region)
    return f"ondemand.{host_region}.saucelabs.com"


REGION_MAP = {
    "us": "us-west-1",
    "eu": "eu-central-1",
    "apac": "apac-southeast-1",
    "headless": "us-east-1",
}



from __future__ import annotations

import contextlib
from typing import Optional

from selenium.webdriver import Remote as WebDriver  # type: ignore

from pytest_sosu.exceptions import WebDriverTestFailed, WebDriverTestInterrupted
from pytest_sosu.logging import get_struct_logger
from pytest_sosu.webdriver.capabilities import Capabilities
from pytest_sosu.webdriver.url import WebDriverUrlData

logger = get_struct_logger(__name__)


@contextlib.contextmanager
def remote_webdriver_ctx(
    url_data: WebDriverUrlData,
    capabilities: Capabilities,
    quit_on_finish: bool = True,
    mark_result_on_finish: bool = True,
    setup_timeouts: bool = True,
):
    wd_safe_url = url_data.to_safe_url()
    logger.debug("Driver starting", capabilities=capabilities, wd_url=wd_safe_url)
    driver = create_remote_webdriver(
        url_data,
        capabilities,
        setup_timeouts=setup_timeouts,
    )
    session_id = driver.session_id
    logger.debug(
        "Driver started",
        capabilities=capabilities,
        wd_url=wd_safe_url,
        driver=driver,
    )
    logger.info("Session started", wd_url=wd_safe_url, session_id=session_id)
    job_result: Optional[str] = "failed"
    try:
        yield driver
        job_result = "passed"
    except WebDriverTestFailed:
        pass
    except (WebDriverTestInterrupted, KeyboardInterrupt):
        job_result = None
    finally:
        if mark_result_on_finish:
            if job_result is not None:
                logger.debug(
                    "Marking test",
                    session_id=session_id,
                    job_result=job_result,
                )
                driver.execute_script(f"sauce:job-result={job_result}")
            else:
                logger.debug(
                    "Not marking test as it was interrupted",
                    session_id=session_id,
                )
        if quit_on_finish:
            logger.debug("Driver quitting", driver=driver)
            driver.quit()
            logger.debug("Driver quitted", driver=driver)
        logger.info("Session stopped", session_id=session_id)


def create_remote_webdriver(
    wd_url_data: WebDriverUrlData,
    capabilities: Capabilities,
    setup_timeouts: bool = True,
) -> WebDriver:
    wd_url = wd_url_data.to_url()
    caps = capabilities.to_dict()
    logger.debug("Dumping caps data", caps=caps)
    wd_safe_url = wd_url_data.to_safe_url()
    logger.debug("Using webdriver URL", wd_url=wd_safe_url)
    driver = WebDriver(
        command_executor=wd_url,
        desired_capabilities=caps,
    )
    if setup_timeouts:
        timeout = capabilities.sauce_options.command_timeout
        if timeout is not None:
            driver.set_page_load_timeout(timeout)
            driver.implicitly_wait(timeout)
            driver.set_script_timeout(timeout)
    return driver



from __future__ import annotations

from enum import Enum
from types import MappingProxyType
from typing import Any, Callable, Dict, Iterator, Mapping, Optional, TypeVar, Union

from pytest_sosu.typing import Literal

_T = TypeVar("_T")
_S = TypeVar("_S")


class DefaultValues(Enum):
    RAISE_EXCEPTION = "raise-exception"


RAISE_EXCEPTION = DefaultValues.RAISE_EXCEPTION


class ImmutableDict(Mapping[_T, _S]):
    def __init__(self, _dict: Mapping[_T, _S]):
        self._mapping_proxy = MappingProxyType(_dict)

    def __getitem__(self, key: _T) -> _S:
        return self._mapping_proxy[key]

    def __iter__(self) -> Iterator[_T]:
        return iter(self._mapping_proxy)

    def __len__(self) -> int:
        return len(self._mapping_proxy)

    def __hash__(self) -> int:
        return hash(tuple(sorted((k, v) for k, v in self._mapping_proxy.items())))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({dict(self._mapping_proxy)})"

    def merge(self, other: ImmutableDict[_T, _S]) -> ImmutableDict[_T, _S]:
        data: Dict[_T, _S] = {}
        data.update(self._mapping_proxy)
        data.update(other)
        return self.__class__(data)


def smart_bool(value: Any) -> bool:
    if isinstance(value, bool):
        return value
    return str(value).lower() in {"true", "t", "yes", "y", "1"}


def convert_or_none(value: Any, converter_func: Callable[[Any], _T]) -> Optional[_T]:
    if value is None:
        return None
    return converter_func(value)


def str_or_none(value: Any) -> Optional[str]:
    return convert_or_none(value, str)


def try_one_of_or_none(
    first: Optional[_T],
    *other_getters: Callable[[], Optional[_T]],
) -> Optional[_T]:
    try:
        return try_one_of(first, *other_getters)
    except ValueError:
        return None


def try_one_of(
    first: Optional[_T],
    *other_getters: Callable[[], Optional[_T]],
    default: Union[_T, Literal[DefaultValues.RAISE_EXCEPTION]] = RAISE_EXCEPTION,
) -> _T:
    value = first
    if value is not None:
        return value
    for value_getter in other_getters:
        value = value_getter()
        if value is not None:
            return value
    if default is RAISE_EXCEPTION:
        raise ValueError("Could not match any of the value")
    return default


def convert_snake_case_to_camel_case(value: str) -> str:
    first_segment, *other_segments = value.split("_")
    return first_segment + "".join(seg.capitalize() for seg in other_segments)



from __future__ import annotations

import dataclasses
import enum
from dataclasses import dataclass
from typing import Any, Dict, Iterator, List, Optional, Sequence, Union

import selenium  # type: ignore
import selenium.webdriver  # type: ignore
import selenium.webdriver.common.by  # type: ignore

from pytest_sosu.logging import get_struct_logger
from pytest_sosu.utils import (
    ImmutableDict,
    convert_snake_case_to_camel_case,
    try_one_of,
    try_one_of_or_none,
)
from pytest_sosu.webdriver.platforms import Browser, Platform

logger = get_struct_logger(__name__)


class SauceTestResultsVisibility(enum.Enum):
    PUBLIC = "public"
    PUBLIC_RESTRICTED = "public restricted"
    SHARE = "share"
    TEAM = "team"
    PRIVATE = "private"


@dataclass(frozen=True)
class SauceOptions:
    name: Optional[str] = None
    build: Optional[str] = None
    tags: Optional[List[str]] = None
    username: Optional[str] = None
    access_key: Optional[str] = None
    custom_data: Optional[Dict[str, Any]] = None
    visibility: Optional[SauceTestResultsVisibility] = None
    tunnel_name: Optional[str] = None
    tunnel_identifier: Optional[str] = None
    tunnel_owner: Optional[str] = None
    parent_tunnel: Optional[str] = None
    record_video: Optional[bool] = None
    video_upload_on_pass: Optional[bool] = None
    record_screenshots: Optional[bool] = None
    record_logs: Optional[bool] = None
    max_duration: Optional[int] = None
    idle_timeout: Optional[int] = None
    command_timeout: Optional[int] = None
    screen_resolution: Optional[str] = None
    extras: ImmutableDict[str, Any] = dataclasses.field(
        default_factory=lambda: ImmutableDict({}),
    )

    auto_include_selenium_version: Optional[bool] = None

    TO_DICT_AUTO_EXCLUDES = (
        "custom_data",
        "visibility",
        "auto_include_selenium_version",
        "extras",
    )

    @classmethod
    def default(cls) -> SauceOptions:
        return cls()

    def __structlog__(self):
        return self.to_dict()

    def merge(self, other: SauceOptions) -> SauceOptions:
        kwargs: Dict[str, Any] = {}
        for field in dataclasses.fields(self):
            name = field.name
            if name == "extras":
                continue
            kwargs[name] = self._merge_field(other, name)
        kwargs["extras"] = self.extras.merge(other.extras)
        new_opts = SauceOptions(**kwargs)
        return new_opts

    def _merge_field(self, other, name):
        return try_one_of_or_none(
            getattr(other, name),
            lambda: getattr(self, name),
        )

    def to_dict(
        self, auto_include_selenium_version: Optional[bool] = None
    ) -> Dict[str, Union[str, int, float]]:

        auto_include_selenium_version = try_one_of(
            auto_include_selenium_version,
            lambda: self.auto_include_selenium_version,
            default=True,
        )

        data = {}
        if auto_include_selenium_version:
            data["seleniumVersion"] = selenium.__version__
        for field in dataclasses.fields(self):
            name = field.name
            if name in self.TO_DICT_AUTO_EXCLUDES:
                continue
            value = getattr(self, name)
            if value is None:
                continue
            dict_name = convert_snake_case_to_camel_case(name)
            data[dict_name] = value
        if self.custom_data is not None:
            data["custom-data"] = self.custom_data
        if self.visibility is not None:
            data["public"] = self.visibility.value
        data.update(self.extras)
        return data


@dataclass(frozen=True)
class Capabilities:
    browser: Optional[Browser] = Browser.default()
    platform: Optional[Platform] = Platform.default()
    sauce_options: SauceOptions = SauceOptions.default()
    extras: ImmutableDict[str, Any] = dataclasses.field(
        default_factory=lambda: ImmutableDict({}),
    )

    w3c_mode: Optional[bool] = None

    @property
    def slug(self):
        if self.browser is None and self.platform is None:
            return "default"
        if self.platform is None:
            return self.browser.slug
        if self.browser is None:
            return self.platform.slug
        return f"{self.browser.slug}-on-{self.platform.slug}"

    def __structlog__(self):
        return self.to_dict()

    def merge(self, other: Capabilities) -> Capabilities:
        new_caps = Capabilities(
            browser=try_one_of_or_none(other.browser, lambda: self.browser),
            platform=try_one_of_or_none(other.platform, lambda: self.platform),
            sauce_options=self.sauce_options.merge(other.sauce_options),
            extras=self.extras.merge(other.extras),
            w3c_mode=try_one_of_or_none(other.w3c_mode, lambda: self.w3c_mode),
        )
        return new_caps

    def to_dict(
        self,
        w3c_mode: Optional[bool] = None,
        auto_include_selenium_version: Optional[bool] = None,
    ) -> Dict[str, Any]:
        w3c_mode = try_one_of(w3c_mode, lambda: self.w3c_mode, default=True)

        data: Dict[str, Any] = {}
        if w3c_mode:
            data.update(
                {
                    "sauce:options": {},
                }
            )
            if self.platform is not None:
                data.update(
                    {
                        "platformName": self.platform.full_name,
                    }
                )
            if self.browser is not None:
                data.update(
                    {
                        "browserName": self.browser.name,
                        "browserVersion": self.browser.version,
                    }
                )
        else:
            if self.platform is not None:
                data.update(
                    {
                        "platform": self.platform.full_name,
                    }
                )
            if self.browser is not None:
                data.update(
                    {
                        "browserName": self.browser.name,
                        "version": self.browser.version,
                    }
                )

        sauce_options_data = data["sauce:options"] if w3c_mode else data
        sauce_options_data.update(
            self.sauce_options.to_dict(
                auto_include_selenium_version=auto_include_selenium_version,
            )
        )

        data.update(self.extras)

        return data


@dataclass(frozen=True)
class CapabilitiesMatrix:
    browsers: Optional[List[Browser]] = None
    platforms: Optional[List[Platform]] = None
    sauce_options_list: Optional[List[SauceOptions]] = None

    def __structlog__(self) -> Dict[str, Any]:
        return self.to_dict()

    def iter_capabilities(self) -> Iterator[Capabilities]:
        browsers: Sequence[Optional[Browser]] = [None]
        if self.browsers is not None:
            browsers = self.browsers
        platforms: Sequence[Optional[Platform]] = [None]
        if self.platforms is not None:
            platforms = self.platforms
        sauce_options_list = (
            [SauceOptions.default()]
            if self.sauce_options_list is None
            else self.sauce_options_list
        )
        for browser in browsers:
            for platform in platforms:
                for sauce_options in sauce_options_list:
                    yield Capabilities(
                        browser=browser,
                        platform=platform,
                        sauce_options=sauce_options,
                    )

    def to_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)



class InvalidMarkerConfiguration(Exception):
    pass


class MultipleMarkerParametersFound(InvalidMarkerConfiguration):
    pass


class WebDriverTestMarkerException(Exception):
    pass


class WebDriverTestFailed(WebDriverTestMarkerException):
    pass


class WebDriverTestInterrupted(WebDriverTestMarkerException):
    pass



# pylint: disable=redefined-outer-name
import datetime
import os
from typing import Any, Callable, Optional

import pytest
from _pytest.config import Config

from pytest_sosu.config import SosuConfig, build_sosu_config
from pytest_sosu.logging import get_struct_logger
from pytest_sosu.plugin_helpers import parametrize_capabilities
from pytest_sosu.webdriver import (
    Browser,
    Capabilities,
    Platform,
    SauceOptions,
    WebDriverTestFailed,
    WebDriverTestInterrupted,
    WebDriverUrlData,
    remote_webdriver_ctx,
)

logger = get_struct_logger(__name__)


def pytest_addoption(parser):
    group = parser.getgroup("sosu plugin sauce labs configuration")

    group.addoption(
        "--sosu-username",
        action="store",
        metavar="SAUCE_USERNAME",
        help="Sauce Labs username",
    )
    group.addoption(
        "--sosu-access-key",
        action="store",
        metavar="SAUCE_ACCESS_KEY",
        help="Sauce Labs access key",
    )
    group.addoption(
        "--sosu-region",
        action="store",
        metavar="SAUCE_REGION",
        help="Sauce Labs region",
    )
    group.addoption(
        "--sosu-webdriver-url",
        action="store",
        metavar="SAUCE_WEBDRIVER_URL",
        help="Sauce Labs WebDriver URL",
    )


def pytest_configure(config: Config):
    logger.debug("pytest_configure", config=config)
    # register an additional marker
    config.addinivalue_line("markers", "sosu(type): mark test to run with Sauce Labs")

    sosu_config = build_sosu_config(config.option, os.environ)
    setattr(config, "sosu", sosu_config)


def _get_sosu_config(config: Config) -> SosuConfig:
    return getattr(config, "sosu")


def pytest_runtest_setup(item: pytest.Item):
    logger.debug("pytest_runtest_setup", item=item)
    sosu_markers = list(item.iter_markers(name="sosu"))
    if sosu_markers:
        logger.debug("sosu marker(s) found", sosu_markers=sosu_markers)


def pytest_generate_tests(metafunc):
    logger.debug(
        "pytest_generate_tests",
        metafunc=metafunc,
        fixturenames=metafunc.fixturenames,
    )

    parametrize_capabilities(metafunc)


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    # Execute all other hooks to obtain the report object.
    outcome = yield
    report = outcome.get_result()

    # Set a report attribute for each phase of a call, which can
    # be "setup", "call", "teardown"
    setattr(item, "report_when_" + report.when, report)


@pytest.fixture(scope="session")
def sosu_build_basename() -> Optional[str]:
    return None


@pytest.fixture(scope="session")
def sosu_build_time_tag() -> str:
    now = datetime.datetime.now()
    date_str = f"{now.year}{now.month:02d}{now.day:02d}"
    time_str = f"{now.hour:02d}{now.minute:02d}"
    return f"{date_str}_{time_str}"


@pytest.fixture(scope="session")
def sosu_build_version(sosu_build_time_tag: str) -> str:
    return sosu_build_time_tag


@pytest.fixture(scope="session")
def sosu_build_name(
    sosu_build_basename: Optional[str], sosu_build_version: str
) -> Optional[str]:
    if sosu_build_basename is None:
        return None
    return f"{sosu_build_basename}_{sosu_build_version}"


@pytest.fixture
def sosu_webdriver_url_data(pytestconfig: Config) -> WebDriverUrlData:
    sosu_config = _get_sosu_config(pytestconfig)
    return sosu_config.webdriver_url_data_with_credentials


@pytest.fixture
def sosu_test_name(request: pytest.FixtureRequest) -> str:
    path = get_function_path(request.node.function)
    return f"{path}::{request.node.name}"


def get_function_path(func: Callable[..., Any]) -> str:
    rel_path = func.__module__.replace(".", "/")
    return f"{rel_path}.py"


@pytest.fixture
def sosu_webdriver_platform() -> Optional[Platform]:
    return Platform.default()


@pytest.fixture
def sosu_webdriver_browser() -> Optional[Browser]:
    return Browser.default()


@pytest.fixture
def sosu_sauce_options(sosu_test_name: str, sosu_build_name: str) -> SauceOptions:
    return SauceOptions(
        name=sosu_test_name,
        build=sosu_build_name,
    )


@pytest.fixture
def sosu_webdriver_capabilities(
    sosu_sauce_options: SauceOptions,
    sosu_webdriver_platform: Optional[Platform],
    sosu_webdriver_browser: Optional[Browser],
) -> Capabilities:
    return Capabilities(
        browser=sosu_webdriver_browser,
        platform=sosu_webdriver_platform,
        sauce_options=sosu_sauce_options,
    )


@pytest.fixture
def sosu_webdriver_parameter_capabilities() -> Capabilities:
    return Capabilities()


@pytest.fixture
def sosu_webdriver_combined_capabilities(
    sosu_webdriver_capabilities: Capabilities,
    sosu_webdriver_parameter_capabilities: Capabilities,
) -> Capabilities:
    return sosu_webdriver_capabilities.merge(sosu_webdriver_parameter_capabilities)


@pytest.fixture
def sosu_webdriver(
    request,
    sosu_webdriver_url_data: WebDriverUrlData,
    sosu_webdriver_combined_capabilities: Capabilities,
):
    with remote_webdriver_ctx(
        sosu_webdriver_url_data,
        sosu_webdriver_combined_capabilities,
    ) as webdriver:
        yield webdriver
        # Using attribute defined in `pytest_runtest_makereport`.
        if not hasattr(request.node, "report_when_call"):
            # No report for test call set - assuming the test was interrupted.
            # Use marker exception for the `remote_webdriver_ctx`.
            raise WebDriverTestInterrupted()
        if request.node.report_when_call.failed:
            # Use marker exception for the `remote_webdriver_ctx`.
            raise WebDriverTestFailed()
